# /// script
# dependencies = [
#   "pydantic-sweep~=0.3.1"
# ]
# ///

import argparse

from config import ExperimentConfig


def command_line_interface() -> ExperimentConfig:
    """Return the configuration passed as json on the command line."""
    # Receive a json-serialized configuration as a command line argument. This
    # requires and equivalent counter-part in the runner.py script.
    parser = argparse.ArgumentParser()
    parser.add_argument("--json", type=str, required=True)
    json_config = parser.parse_args().json

    # Reconstruct the config from json
    return ExperimentConfig.model_validate_json(json_config)


def main(config: ExperimentConfig):
    """Main training function."""
    print(f"Execute main with: {config!r}")
    # Your favorite program goes here...


if __name__ == "__main__":
    config = command_line_interface()
    main(config)



from __future__ import annotations

from collections.abc import Hashable, Iterable
from typing import Protocol, TypeAlias, TypeVar, Union

import pydantic

__all__ = [
    "Chainer",
    "Combiner",
    "Config",
    "FieldValue",
    "FlexibleConfig",
    "Path",
    "StrictPath",
]


StrictPath: TypeAlias = tuple[str, ...]
"""A tuple-path of keys for a pydantic model."""

Path: TypeAlias = Union[Iterable[str], str, "StrictPath"]
"""Anything that can be converted to a tuple-path (str or iterable of str)."""

FieldValue: TypeAlias = Hashable | pydantic.BaseModel
"""The possible values that should be assigned to a field.

Fields should be hashable (and therefore immutable) values. That makes them safer to 
use in a configuration, since unlike mutable types they can not be modified inplace.
"""

Config: TypeAlias = dict[str, Union["FieldValue", "Config"]]
"""A nested config dictionary for configurations."""

FlexibleConfig: TypeAlias = Union[
    dict["Path", Union["FieldValue", "FlexibleConfig"]], "Config"
]
"""A flexible config that allows any Path."""

T = TypeVar("T")


class Combiner(Protocol[T]):
    """A function that yields tuples of items."""

    def __call__(self, *configs: Iterable[T]) -> Iterable[tuple[T, ...]]: ...


class Chainer(Protocol[T]):
    """A function that chains iterables together."""

    def __call__(self, *configs: Iterable[T]) -> Iterable[T]: ...


BaseModelT = TypeVar("BaseModelT", bound=pydantic.BaseModel)
"""TypeVar for a pydantic BaseModel."""



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import importlib.util
import sys
from pathlib import Path

PROJECT_ROOT = Path(__file__).parents[1]
MODULE_ROOT = PROJECT_ROOT.joinpath("src", "pydantic_sweep")

sys.path.append(str(PROJECT_ROOT))


def import_file(file: Path, /):
    """Import a file directly."""
    assert file.exists()
    spec = importlib.util.spec_from_file_location(file.stem, file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "pydantic_sweep"
author = "Felix Berkenkamp"
release = import_file(MODULE_ROOT / "_version.py").__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",  # Docstrings
    "sphinx_rtd_theme",  # Theme
    "sphinx.ext.autodoc",  # Source code documentation
    "sphinx.ext.intersphinx",  # Link to external documentation
    "autoapi.extension",  # Automatic API generation
    "sphinx.ext.linkcode",  # Link to Github code
    "myst_nb",  # Executable notebooks
    "sphinx_copybutton",  # Copy code cells
]

nitpicky = True
default_role = "any"
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Add links to other libraries
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "more_itertools": ("https://more-itertools.readthedocs.io/en/stable", None),
    "pydantic": ("https://docs.pydantic.dev/latest/", None),
}

nitpick_ignore = [
    ("py:class", "collections.abc.Iterable"),
    ("py:obj", "pydantic.BaseModel"),
    ("py:class", "pydantic.BaseModel"),
    ("py:class", "T"),
    ("py:obj", "T"),
    ("py:class", "Ellipsis"),
    # These are somehow not found
    ("py:class", "pydantic_sweep.types.Config"),
    ("py:class", "pydantic_sweep.types.Path"),
    ("py:class", "pydantic_sweep.types.StrictPath"),
    ("py:class", "pydantic_sweep.types.FieldValue"),
    ("py:class", "pydantic_sweep.types.BaseModelT"),
    ("py:class", "pydantic_sweep.types.FlexibleConfig"),
]
napoleon_use_rtype = False

autoapi_dirs = ["../src/pydantic_sweep"]
autoapi_options = [
    "members",
    "undoc-members",
    # "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]
autoapi_root = "autoapi"
autoapi_keep_files = False
keep_warnings = True
autodoc_typehints = "signature"

# Notebook formatting
source_suffix = {
    ".rst": "restructuredtext",
    ".ipynb": "myst-nb",
    ".myst": "myst-nb",
}
# Note: Multi-suffixes like .pct.py will be supported in myst-nb>1.1.2
nb_custom_formats = {
    ".pctpy": ["jupytext.reads", {"fmt": "py:percent"}],
}
nb_execution_in_temp = True
nb_execution_allow_errors = False


def linkcode_resolve(domain, info):
    """Link to code on Github."""
    # https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html#module-sphinx.ext.linkcode
    if domain != "py":
        return None
    if not info["module"]:
        return None
    # linkcode is not aware of import in __init__, so need to do some annoying
    # duplication here. Could perhaps be replaced by importing each function and
    # checking `__module__`.
    filename = info["module"].replace(".", "/")
    py_obj_name = info["fullname"]
    if filename.endswith(".types"):
        filename = f"{filename}.py"
    elif py_obj_name == "random_seeds":
        filename = f"{filename}/_utils.py"
    elif py_obj_name == "__version__":
        filename = f"{filename}/_version.py"
    else:
        filename = f"{filename}/_model.py"
    return f"https://github.com/befelix/pydantic_sweep/blob/main/src/{filename}"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
# https://docs.readthedocs.io/en/stable/guides/edit-source-links-sphinx.html
html_context = {
    "display_github": True,
    "github_user": "befelix",
    "github_repo": "pydantic_sweep",
    "github_version": "main",
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}
# Do not copy the source code files into the documentation
html_copy_source = False



import re
import subprocess
import sys
from pathlib import Path

ROOT = Path(__file__).parents[1]
README = ROOT / "README.md"


def extract_python_code(markdown_text):
    pattern = r"```python\n(.*?)\n```"
    return re.findall(pattern, markdown_text, re.DOTALL)


def test_extract_markdown():
    FUN1 = """\
 def hello():
    print("hi")   
"""
    FUN2 = """\
x = 10
y = 20
print(x + y)
"""
    markdown_content = f"""\
Here is some text.

```python
{FUN1}
```

Some more text.

```python
{FUN2}
```
"""
    extracted = extract_python_code(markdown_content)
    assert extracted == [FUN1, FUN2]


def test_readme():
    """Make sure the code in the readme can be executed."""
    with open(README) as f:
        content = f.read()

    code = "\n".join(extract_python_code(content))
    subprocess.run([sys.executable, "-c", code], check=True, capture_output=True)



import collections.abc
import copy
import dataclasses
import typing
from typing import Annotated, Any, TypeVar

import pydantic
import pytest
import typing_extensions

from pydantic_sweep._utils import (
    RaiseWarnIgnore,
    _flexible_config_to_nested,
    as_hashable,
    iter_subtypes,
    merge_nested_dicts,
    model_diff,
    nested_dict_at,
    nested_dict_from_items,
    nested_dict_get,
    nested_dict_replace,
    normalize_path,
    raise_warn_ignore,
    random_seeds,
)


def test_normalize_path():
    path = ("a", "A_", "b0", "__C")
    assert normalize_path(path) == path
    assert normalize_path("a.A_.b0.__C") == path

    with pytest.raises(ValueError):
        normalize_path("a,b")
    with pytest.raises(ValueError):
        normalize_path(".")
    with pytest.raises(ValueError):
        normalize_path("a.b.")
    with pytest.raises(ValueError):
        normalize_path("a..b")
    with pytest.raises(ValueError):
        normalize_path(".a.b")

    with pytest.raises(ValueError):
        normalize_path(("a", "2"), check_keys=True)
    with pytest.raises(ValueError):
        normalize_path(("a.b",), check_keys=True)
    with pytest.raises(ValueError):
        normalize_path(("0a.b",), check_keys=True)


class TestNormalizeFlexibleConfig:
    def test_basic(self):
        assert _flexible_config_to_nested(dict(a=1, b=2)) == dict(a=1, b=2)
        assert _flexible_config_to_nested({"a.b": 1}) == dict(a=dict(b=1))
        assert _flexible_config_to_nested({("a", "b"): 1}) == dict(a=dict(b=1))
        assert _flexible_config_to_nested(dict(a={"b.c": 1})) == dict(
            a=dict(b=dict(c=1))
        )

    def test_nested(self):
        assert _flexible_config_to_nested({"a.b.c": 1, "a.b.d": 2}) == dict(
            a=dict(b=dict(c=1, d=2))
        )
        assert _flexible_config_to_nested({"a.b.c": 1, "a.b.d.e": 2}) == dict(
            a=dict(b=dict(c=1, d=dict(e=2)))
        )

    def test_conflicts(self):
        with pytest.raises(ValueError):
            _flexible_config_to_nested({"a.b": 1, "a": 2})
        with pytest.raises(ValueError):
            _flexible_config_to_nested({"a": dict(b=1), "a.b": 2})
        with pytest.raises(ValueError):
            _flexible_config_to_nested({"a.b.c": 1, "a.b.c.d": 2})

    def test_skip(self):
        assert _flexible_config_to_nested({"a.c": None, "a.b": 2}, skip=None) == dict(
            a=dict(b=2)
        )


class TestNestedDictFromItems:
    def test_functionality(self):
        d = {("a", "a"): 5, ("a", "b", "c"): 6, "c": 7}
        res = dict(a=dict(a=5, b=dict(c=6)), c=7)
        assert nested_dict_from_items(d.items()) == res

    def test_duplicate_key(self):
        with pytest.raises(ValueError):
            nested_dict_from_items([("a", 1), ("a", 1)])
        with pytest.raises(ValueError):
            nested_dict_from_items([("a.a", 1), ("a.a", 1)])

    def test_parent_overwrite(self):
        with pytest.raises(ValueError):
            nested_dict_from_items([("a.a", 5), ("a", 6)])
        with pytest.raises(ValueError):
            nested_dict_from_items([("a.a.a", 5), ("a", 6)])
        with pytest.raises(ValueError):
            nested_dict_from_items([("a.a.a", 5), ("a.a", 6)])

    def test_child_overwrite(self):
        with pytest.raises(ValueError):
            nested_dict_from_items([("a", 6), ("a.a", 5)])
        with pytest.raises(ValueError):
            nested_dict_from_items([("a", 6), ("a.a", 5)])
        with pytest.raises(ValueError):
            nested_dict_from_items([("a.a", 6), ("a.a.a", 5)])


def test_nested_dict_at():
    res = nested_dict_at("a.b.c", 5)
    assert res == dict(a=dict(b=dict(c=5)))


def test_nested_dict_replace():
    d = dict(a=5, b=dict(c=6, d=7))
    d_orig = copy.deepcopy(d)
    expected = dict(a=5, b=dict(c=0, d=7))

    res = nested_dict_replace(d, "b.c", value=0)
    assert res == expected
    assert d == d_orig, "In-place modification"


class TestNestedDictGet:
    def test_basic(self):
        d = dict(a=dict(b=dict(c=5)))

        assert nested_dict_get(d, "a") is d["a"]
        assert nested_dict_get(d, "a.b") is d["a"]["b"]
        assert nested_dict_get(d, "a.b.c") == 5

        with pytest.raises(KeyError):
            nested_dict_get(d, "c")

    def test_leaf(self):
        d = dict(a=dict(b=2))
        nested_dict_get(d, "a", leaf=False)
        nested_dict_get(d, "a.b", leaf=True)

        with pytest.raises(ValueError):
            nested_dict_get(d, "a", leaf=True)
        with pytest.raises(ValueError):
            nested_dict_get(d, "a.b", leaf=False)


def test_merge_dicts():
    res = dict(a=dict(a=5, b=dict(c=6, y=9)), c=7)
    d1 = dict(a=dict(a=5, b=dict(c=6)))
    d2 = dict(c=7, a=dict(b=dict(y=9)))
    assert merge_nested_dicts(d1, d2) == res

    # This is already tested as part of TestUnflattenItems
    with pytest.raises(ValueError):
        merge_nested_dicts(dict(a=1), dict(a=2))
    with pytest.raises(ValueError):
        merge_nested_dicts(dict(a=dict(a=5)), dict(a=6))
    with pytest.raises(ValueError):
        merge_nested_dicts(dict(a=6), dict(a=dict(a=5)))

    assert merge_nested_dicts(dict(a=1), dict(b=2), overwrite=True) == dict(a=1, b=2)
    assert merge_nested_dicts(dict(a=1), dict(a=2), overwrite=True) == dict(a=2)
    assert merge_nested_dicts(dict(a=dict(b=2)), dict(a=3), overwrite=True) == dict(a=3)


class TestAsHashable:
    def test_builtins(self):
        hash(as_hashable(None)) == hash(None)
        hash(as_hashable((1, 2))) == hash((1, 2))

    def test_dict(self):
        hash(as_hashable(dict(a=dict(b=2))))
        assert as_hashable(dict(a=1, b=2)) == as_hashable(dict(b=2, a=1))

        res1 = as_hashable(dict(a=1, b=dict(c=2)))
        res2 = as_hashable(dict(b=dict(c=2), a=1))
        assert res1 == res2

        assert as_hashable(dict(a=1)) != as_hashable(dict(a=2))
        assert as_hashable(dict(a=dict(b=1))) != as_hashable(dict(b=dict(a=1)))
        assert as_hashable([1, 2]) != as_hashable((1, 2))

    def test_pydantic(self):
        class Sub(pydantic.BaseModel):
            x: int

        class Model(pydantic.BaseModel):
            a: int
            sub: Sub

        hash(as_hashable(dict(a=1, b=dict(c=2))))
        hash(as_hashable(dict(a=1, b=Model(a=1, sub=Sub(x=1)))))

        class Model(pydantic.BaseModel):
            a: int
            b: int

        assert as_hashable(Model(a=1, b=2)) == as_hashable(Model(b=2, a=1))

        class Model1(pydantic.BaseModel):
            x: int

        class Model2(pydantic.BaseModel):
            x: int

        assert as_hashable(Model1(x=1)) != as_hashable(Model2(x=1))
        assert as_hashable(Model1(x=1)) != as_hashable(dict(x=1))

    def test_set(self):
        hash(as_hashable(set([1, 2])))
        assert as_hashable({1, 2}) == as_hashable({1, 2})
        assert as_hashable({1, 2}) != as_hashable((1, 2))

    def test_exception(self):
        @dataclasses.dataclass
        class Test:
            x: int

        t = Test(x=5)

        with pytest.raises(TypeError):
            as_hashable(t)


def test_random_seeds():
    assert set(random_seeds(10, upper=10)) == set(range(10))
    with pytest.raises(ValueError):
        random_seeds(-1)
    with pytest.raises(ValueError):
        random_seeds(1, upper=0)


def test_raise_warn_ignore():
    class CustomException(Exception):
        pass

    class CustomWarning(UserWarning):
        pass

    raise_warn_ignore("blub", action="ignore")
    with pytest.raises(CustomException, match="blub1"):
        raise_warn_ignore("blub1", action="raise", exception=CustomException)
    with pytest.warns(CustomWarning, match="blub2"):
        raise_warn_ignore("blub2", action="warn", warning=CustomWarning)

    with pytest.raises(ValueError, match="raise, warn, ignore"):
        raise_warn_ignore("blub", action="OWEH")

    raise_warn_ignore("blub", action=RaiseWarnIgnore.IGNORE)


def subtypes(x) -> set:
    return set(iter_subtypes(x))


class TestIterSubtypes:
    def test_basic(self):
        assert subtypes(float) == {float}
        assert subtypes(int) == {int}
        assert subtypes(None) == {None}

        class Test(list):
            """Custom type"""

        assert subtypes(Test) == {Test}
        assert subtypes(collections.abc.Sequence[str]) == {
            collections.abc.Sequence,
            str,
        }

    def test_old(self):
        assert subtypes(typing.Dict) == {dict}  # noqa: UP006
        assert subtypes(typing.List) == {list}  # noqa: UP006
        assert subtypes(typing.List[typing.Dict]) == {list, dict}  # noqa: UP006
        assert subtypes(typing.Sequence[str]) == {collections.abc.Sequence, str}  # noqa: UP006, RUF100

    def test_final(self):
        assert subtypes(typing.Final[str]) == {str}
        assert subtypes(typing_extensions.Final[str]) == {str}

    def test_generic(self):
        T = TypeVar("T", bound=str | int)
        assert subtypes(T) == {str, int}
        T = TypeVar("T", str, int)
        assert subtypes(T) == {str, int}
        assert subtypes(T | float) == {str, int, float}
        T = TypeVar("T")
        assert subtypes(T) == {Any}

    def test_alias(self):
        assert subtypes(list[str]) == {list, str}
        assert subtypes(list[str | float]) == {list, str, float}
        assert subtypes(list[str | float]) == {list, str, float}
        assert subtypes(set[tuple[str] | tuple[float, ...]]) == {set, tuple, str, float}

    def test_generic_alias(self):
        T = TypeVar("T")
        assert subtypes(list[T]) == {list, typing.Any}
        T = TypeVar("T", str, float)
        assert subtypes(list[T]) == {list, str, float}

    def test_annotated(self):
        assert subtypes(Annotated[int, "blub"]) == {int}
        assert subtypes(Annotated[int | Annotated[float, "s"], "blub"]) == {int, float}

    def test_literal(self):
        assert subtypes(typing.Literal["test", 1]) == {str, int}
        assert subtypes(typing_extensions.Literal["test", 1]) == {str, int}


class TestModelDiff:
    def test_none(self):
        assert not model_diff(None, None)

    def test_basic(self):
        class Sub(pydantic.BaseModel):
            x: int = 0
            y: int = 0

        class Model(pydantic.BaseModel):
            sub: Sub = Sub()
            z: str = "hi"

        assert not model_diff(Model(), Model())
        assert model_diff(Sub(), Sub(x=1)) == dict(x=(0, 1))
        assert model_diff(Sub(), Sub(x=1, y=2)) == dict(x=(0, 1), y=(0, 2))
        assert model_diff(Model(), Model(sub=Sub(x=1))) == dict(sub=dict(x=(0, 1)))

    def test_different_models(self):
        class M(pydantic.BaseModel):
            x: int = 0

        class Y(pydantic.BaseModel):
            x: int = 0

        with pytest.raises(ValueError):
            model_diff(M(), Y())

    def test_unhashable(self):
        class Model(pydantic.BaseModel):
            x: list | tuple

        # unhashable
        assert not model_diff(Model(x=[1]), Model(x=[1]))
        assert model_diff(Model(x=(1,)), Model(x=[1])) == dict(x=((1,), [1]))

    def test_different_submodules(self):
        class S1(pydantic.BaseModel):
            x: int = 0

        class S2(pydantic.BaseModel):
            x: int = 0

        class Model(pydantic.BaseModel):
            sub: S1 | S2

        assert model_diff(Model(sub=S1()), Model(sub=S2())) == dict(sub=(S1(), S2()))

    def test_list(self):
        x1 = [0, 1, 2]
        x2 = [0, 99, 2]
        assert model_diff(x1, x2) == {"[1]": (1, 99)}

    def test_dict(self):
        d1 = dict(a="a", b="b", c="c")
        d2 = dict(c="c", b="b", a=0)
        assert model_diff(d1, d2) == {"[a]": ("a", 0)}



from __future__ import annotations

import copy
import enum
import itertools
import operator
import random
import re
import types
import typing
import warnings
from collections.abc import Callable, Hashable, Iterable, Iterator, Mapping, Sequence
from typing import Any, Literal, TypeVar, overload

import pydantic
from typing_extensions import Self

from pydantic_sweep.types import Config, FieldValue, FlexibleConfig, Path, StrictPath

__all__ = [
    "as_hashable",
    "items_skip",
    "merge_nested_dicts",
    "model_diff",
    "nested_dict_at",
    "nested_dict_from_items",
    "nested_dict_get",
    "nested_dict_items",
    "nested_dict_replace",
    "normalize_path",
    "notebook_link",
    "path_to_str",
    "raise_warn_ignore",
    "random_seeds",
]


T = TypeVar("T")
GenericUnion = type(T | str)
"""A Generic Union type"""

SpecialGenericAlias = type(typing.List[str])  # noqa: UP006
"""Old-style GenericAlias"""

valid_key_pattern = r"[A-Za-z_][A-Za-z0-9_]*"
# Valid python keys starts with letters and can contain numbers and underscores after
_STR_PATH_PATTERN = re.compile(rf"^{valid_key_pattern}(\.{valid_key_pattern})*$")
_STR_KEY_PATTERN = re.compile(rf"^{valid_key_pattern}$")
# Explanation:
# ^ - Matches the start of the string.
# We first match a valid key, followed by dot-seperated valid keys
# $ - Matches the end of the string.


def path_to_str(p: Path, /) -> str:
    return p if isinstance(p, str) else ".".join(p)


def normalize_path(path: Path, /, *, check_keys: bool = False) -> StrictPath:
    """Normalize a path to a tuple of strings.

    Parameters
    ----------
    path :
        The path to be normalized.
    check_keys :
        If ``True``, also check each individual key in a tuple path.
    """
    match path:
        case str():
            if not re.fullmatch(_STR_PATH_PATTERN, path):
                raise ValueError(
                    "If provided as a string, the path must consist only of "
                    f"dot-separated keys. For example, 'my.key'. Got {path})"
                )
            return tuple(path.split("."))
        case tuple():
            pass
        case Iterable():
            path = tuple(path)
        case _:
            raise ValueError(f"Expected a path, got {path}")

    if check_keys:
        for p in path:
            if not re.fullmatch(_STR_KEY_PATTERN, p):
                raise ValueError(
                    f"Paths can only contain letters and underscores, got {p}."
                )

    return path


@overload
def nested_dict_get(d: Config, /, path: Path, *, leaf: Literal[True]) -> FieldValue: ...


@overload
def nested_dict_get(d: Config, /, path: Path, *, leaf: Literal[False]) -> Config: ...


def nested_dict_get(
    d: Config, /, path: Path, *, leaf: bool | None = None
) -> Config | FieldValue:
    """Return the value of a nested dict at a certain path.

    Parameters
    ----------
    d
        The config to check
    path
        A path that we want to resolve.
    leaf
        If ``True``, check that we return a leaf node. If ``False``, check that we
        return a non-leaf node.

    Raises
    ------
    AttributeError
        If the path does not exist.
    ValueError
        If the result does not match what we specified in ``leaf``.
    """
    path = normalize_path(path)
    for p in path:
        d = d[p]  # type: ignore[assignment]
    if leaf is not None:
        if leaf:
            if isinstance(d, dict):
                raise ValueError(
                    f"Expected a leaf at path {path_to_str(path)}, but got a "
                    f"dictionary."
                )
        else:
            if not isinstance(d, dict):
                raise ValueError(
                    f"Expected a non-leaf node at path {path_to_str(path)}, but got "
                    f"{type(d)}."
                )
    return d


def nested_dict_replace(
    d: Config, /, path: Path, value: FieldValue, *, inplace: bool = False
) -> Config:
    """Replace the value of a nested dict at a certain path (out of place)."""
    if not inplace:
        d = copy.deepcopy(d)

    *subpath, final = normalize_path(path)

    node = d
    for i, key in enumerate(subpath):
        sub = node[key]
        if not isinstance(sub, dict):
            raise ValueError(
                f"Expected a dictionary at {path_to_str(subpath[: i + 1])}, got {sub}."
            )
        node = sub

    if final not in node:
        raise KeyError(f"The path '{path_to_str(path)}' is not part of the dictionary.")
    else:
        node[final] = value

    return d


def nested_dict_at(path: Path, value: FieldValue) -> Config:
    """Return nested dictionary with the value at path."""
    path = normalize_path(path)
    return nested_dict_from_items([(path, value)])


def nested_dict_from_items(
    items: Iterable[tuple[StrictPath, FieldValue | Config]], /
) -> Config:
    """Convert paths and values (items) to a nested dictionary.

    Paths are assumed as single dot-separated strings.
    """
    result: dict[str, Any] = dict()

    for full_path, value in items:
        *path, key = full_path
        node = result

        for part in path:
            if part not in node:
                node[part] = dict()

            node = node[part]

            if not isinstance(node, dict):
                raise ValueError(
                    f"In the configs, for '{path_to_str(path)}' there are both a "
                    f"value ({node}) and child nodes with values defined. "
                    "This means that these two configs would overwrite each other."
                )

        if key in node:
            if isinstance(node[key], dict):
                raise ValueError(
                    f"In the configs, for '{path_to_str(full_path)}' there are both a"
                    f" value ({value}) and child nodes with values defined. "
                    "This means that these two configs would overwrite each other."
                )
            else:
                raise ValueError(
                    f"The key {path_to_str(full_path)} has conflicting values "
                    f"assigned: {node[key]} and {value}."
                )
        else:
            node[key] = value

    return result


def _nested_dict_items(
    d: FlexibleConfig, /, path: StrictPath
) -> Iterator[tuple[StrictPath, FieldValue]]:
    """See nested_dict_items"""
    if not isinstance(d, dict):
        raise ValueError(f"Expected a dictionary, got {d} of type {type(d)}.")
    for subkey, value in d.items():
        cur_path = (*path, *normalize_path(subkey))
        if isinstance(value, dict):
            yield from _nested_dict_items(value, path=cur_path)
        else:
            yield cur_path, value


def nested_dict_items(d: FlexibleConfig, /) -> Iterator[tuple[StrictPath, FieldValue]]:
    """Yield paths and leaf values of a nested dictionary.

    Note: This function has special handling for Paths, i.e., it will expand
    dot-separated keys and tuples-keys into nested paths.

    >>> list(nested_dict_items(dict(a=dict(b=3), c=2)))
    [(('a', 'b'), 3), (('c',), 2)]
    """
    return _nested_dict_items(d, path=())


def merge_nested_dicts(*dicts: FlexibleConfig, overwrite: bool = False) -> Config:
    """Merge multiple Config dictionaries into a single one.

    This function includes error checking for duplicate keys and accidental overwriting
    of subtrees in the nested configuration objects.

    >>> merge_nested_dicts(dict(a=dict(b=2)), dict(c=3))
    {'a': {'b': 2}, 'c': 3}

    >>> merge_nested_dicts(dict(a=dict(b=2)), dict(a=5), overwrite=True)
    {'a': 5}
    """
    if not overwrite:
        return nested_dict_from_items(
            itertools.chain(*(nested_dict_items(d) for d in dicts))
        )

    res: Config = dict()
    for d in dicts:
        for path, value in nested_dict_items(d):
            node = res
            *subpath, final = path
            for p in subpath:
                if p not in node or not isinstance(node[p], dict):
                    node[p] = dict()
                node = node[p]  # type: ignore[assignment]
            node[final] = value

    return res


class _NoSkip:
    pass


def _flexible_config_to_nested(
    config: FlexibleConfig, /, skip: Any = _NoSkip
) -> Config:
    """Normalize a flexible config to a nested dictionary."""
    items = nested_dict_items(config)
    if skip is not _NoSkip:
        items = items_skip(items, target=skip)
    return nested_dict_from_items(items)


K = TypeVar("K")
V = TypeVar("V")


def items_skip(items: Iterable[tuple[K, V]], target: Any) -> Iterator[tuple[K, V]]:
    """Yield items skipping certain targets."""
    for key, value in items:
        if value is not target:
            yield key, value


def as_hashable(item: Any, /) -> Hashable:
    """Convert input into a unique, hashable representation.

    In addition to the builtin hashable types, this also works for dictionaries and
    pydantic Models (recursively). Sets and lists are also supported, but not dealt
    with recursively.

    Parameters
    ----------
    item
        The item that we want to convert to something hashable.

    Returns
    -------
    Hashable
        A hashable representation of the item.
    """
    match item:
        case Hashable():
            return item
        case pydantic.BaseModel():
            # Cannot use model_dump here, since that would map different models with
            # the same attribute to the same keys. Don't need frozenset, since key
            # order is deterministic
            model_dump = tuple(
                (key, as_hashable(getattr(item, key)))
                for key in item.__class__.model_fields
            )
            return f"pydantic:{item.__class__}:{model_dump}"
        case dict():
            return frozenset((key, as_hashable(value)) for key, value in item.items())
        case set():
            return frozenset(item)
        case list():
            return ("__list_type", tuple(item))
        case _:
            raise TypeError(f"Unhashable object of type {type(item)}")


def random_seeds(num: int, *, upper: int = 1000) -> list[int]:
    """Generate unique random values within a certain range.

    This is useful in scenarios where we don't want to hard-code a random seed,
    but also need reproducibility by setting a seed. Sampling the random seed is a
    good compromise there.

    Parameters
    ----------
    num :
        The number of random seeds to generate.
    upper:
        A non-inclusive upper bound on the maximum seed to generate.

    Returns
    -------
    list[int]:
        A list of integer seeds.
    """
    if upper <= 0:
        raise ValueError("Upper bound must be positive.")

    return random.sample(range(upper), num)


def notebook_link(
    name: Literal["combinations", "example", "intro", "models", "nested"],
    *,
    version: Literal["stable", "latest"] = "stable",
) -> str:
    return f"https://pydantic-sweep.readthedocs.io/{version}/notebooks/{name}.html"


class RaiseWarnIgnore(enum.Enum):
    """Actions for `raise_warn_ignore`."""

    RAISE = "raise"
    WARN = "warn"
    IGNORE = "ignore"

    @classmethod
    def cast(cls, name: str | Self, /) -> Self:
        try:
            return cls(name)
        except ValueError:
            options = ", ".join([action.value for action in cls])
            raise ValueError(f"{name} is not a valid action. Options are: {options}")


def raise_warn_ignore(
    message: str,
    *,
    action: Literal["raise", "warn", "ignore"] | RaiseWarnIgnore,
    exception: type[Exception] = ValueError,
    warning: type[Warning] = UserWarning,
) -> None:
    """Raise/warn/ignore depending on action input."""
    action = RaiseWarnIgnore.cast(action)
    if action is RaiseWarnIgnore.WARN:
        warnings.warn(message, category=warning)
    elif action is RaiseWarnIgnore.RAISE:
        raise exception(message)


def iter_subtypes(t: type, /) -> Iterator[type]:
    """Iterate over all possible subtypes of the input type.

    >>> list(iter_subtypes(str | int))
    [<class 'str'>, <class 'int'>]

    >>> T = TypeVar("T", bound=str)
    >>> list(iter_subtypes(T | float | int))
    [<class 'str'>, <class 'float'>, <class 'int'>]
    """
    origin = typing.get_origin(t)

    match (origin, t):
        case (typing.Annotated, _):
            sub = typing.get_args(t)[0]
            yield from iter_subtypes(sub)
        case (typing.Union | types.UnionType, _):
            for arg in typing.get_args(t):
                yield from iter_subtypes(arg)
        case (typing.Final, _):
            yield from iter_subtypes(*typing.get_args(t))
        case (typing.Literal, _):
            for arg in typing.get_args(t):
                yield type(arg)
        case (_, types.GenericAlias() | SpecialGenericAlias()):  # type: ignore
            # Generic alias: list[str], special: typing.List[str]
            if origin is not None:
                yield from iter_subtypes(origin)
            for arg in typing.get_args(t):
                if arg is not Ellipsis:
                    yield from iter_subtypes(arg)
        case (_, typing.TypeVar()):
            if t.__bound__ is not None:
                yield from iter_subtypes(t.__bound__)
            elif t.__constraints__:
                for constraint in t.__constraints__:
                    yield from iter_subtypes(constraint)
            else:
                # Unconstrained typevar can take any value.
                yield typing.Any
        case _:
            if origin is None:
                yield t
            else:
                # Resolves special types like typing.List
                yield origin


class MissingMeta(type):
    def __repr__(cls) -> str:
        return "Missing"


class Missing(metaclass=MissingMeta):
    """A Missing value"""

    pass


def _model_diff_iter(
    m1: Any, m2: Any, *, path: StrictPath = (), compare: Callable[[Any, Any], bool]
) -> Iterator[tuple[StrictPath, tuple]]:
    """Iterator implementation for model_diff."""
    # Different types are treated differently by design. One could in principle
    # compare tuples and dicts, but given that the core usecase is pydantic models
    # these will get normalized to the same type in any case.
    cls = type(m1)
    if cls is not type(m2):
        yield path, (m1, m2)
        return

    match m1:
        case pydantic.BaseModel():
            # Thanks to previous check, we know they have the same keys
            for name in cls.model_fields:
                value1 = getattr(m1, name)
                value2 = getattr(m2, name)
                yield from _model_diff_iter(
                    value1, value2, path=(*path, name), compare=compare
                )

        case Mapping():
            keys = m1.keys() | m2.keys()
            for key in keys:
                if key not in m1:
                    yield (*path, f"[{key}]"), (Missing, m2[key])
                elif key not in m2:
                    yield (*path, f"[{key}]"), (m1[key], Missing)
                else:
                    yield from _model_diff_iter(
                        m1[key], m2[key], path=(*path, f"[{key}]"), compare=compare
                    )

        case Sequence() if not isinstance(m1, str):
            for i, (v1, v2) in enumerate(
                itertools.zip_longest(m1, m2, fillvalue=Missing)
            ):
                yield from _model_diff_iter(
                    v1, v2, path=(*path, f"[{i}]"), compare=compare
                )

        case _:
            # A leaf node that we cannot expand directly --> call standard comparator
            if not compare(m1, m2):
                yield path, (m1, m2)


def model_diff(
    m1: Any,
    m2: Any,
    /,
    *,
    compare: Callable[[Any, Any], bool] = operator.eq,
) -> dict[str, Any]:
    """Return a nested dictionary of model diffs.

    That is, given two models this function iterates over all nested sub-model fields
    and returns a nested dictionaries with leaf values that are tuples of the
    corresponding value of the left model, and the value of the right model.

    Parameters
    ----------
    m1 :
        The first model, or nested structure.
    m2 :
        The second model, or nested structure
    compare :
        Function to compare two elements, returns ``True`` if they are equal.

    Examples
    --------
    >>> class Sub(pydantic.BaseModel):
    ...     x: int = 0

    >>> class Model(pydantic.BaseModel):
    ...     s: Sub = Sub()
    ...     y: int = 2

    >>> model_diff(Model(s=Sub(x=1)), Model(s=Sub(x=2)))
    {'s': {'x': (1, 2)}}

    >>> model_diff(Model(), Model())
    {}
    """
    return nested_dict_from_items(_model_diff_iter(m1, m2, compare=compare))



"""Larger integration tests for the overall workflow."""

import pydantic_sweep as ps
from pydantic_sweep import field


def test_basic_class():
    """Basic workflow using a BaseModel."""

    class Model(ps.BaseModel):
        x: int = 6
        y: int = 5
        z: str = "a"

    configs = ps.initialize(
        Model,
        ps.config_product(
            ps.field("z", [ps.DefaultValue, "b"]),
            ps.config_zip(
                ps.field("x", [1, 2]),
                ps.field("y", [3, 4]),
            ),
        ),
    )
    expected = [
        Model(x=1, y=3),
        Model(x=2, y=4),
        Model(x=1, y=3, z="b"),
        Model(x=2, y=4, z="b"),
    ]
    assert configs == expected


def test_basic_instance():
    """Basic workflow using an instance of a model."""

    class Model(ps.BaseModel):
        x: int = 6
        y: int = 5
        z: str = "a"

    configs = ps.initialize(
        Model,
        ps.config_product(
            ps.field("z", [ps.DefaultValue, "b"]),
            ps.config_zip(
                ps.field("x", [1, 2]),
                ps.field("y", [3, 4]),
            ),
        ),
        default=dict(z="new"),
    )
    expected = [
        Model(x=1, y=3, z="new"),
        Model(x=2, y=4, z="new"),
        Model(x=1, y=3, z="b"),
        Model(x=2, y=4, z="b"),
    ]
    assert configs == expected


def test_nested():
    """Test nested model configuration."""

    class Sub1(ps.BaseModel):
        x: int = 5
        z: int = 7

    class Sub2(ps.BaseModel):
        y: int = 6
        z: int = 9

    class Model(ps.BaseModel):
        s: Sub2 | Sub1

    models = ps.initialize(
        Model,
        ps.config_chain(
            ps.initialize(Sub1, field("x", [1, 2]), to="s"),
            ps.initialize(Sub2, field("y", [3]), to="s"),
        ),
    )
    assert models == [Model(s=Sub1(x=1)), Model(s=Sub1(x=2)), Model(s=Sub2(y=3))]



import re
import subprocess
import sys
from pathlib import Path

from pydantic_sweep import __version__

_EXPERIMENT_DIR = Path(__file__).parents[1] / "example"


def test_script_version():
    for name in ["runner.py", "train.py"]:
        with open(_EXPERIMENT_DIR / name) as f:
            for line in f.readlines():
                # No longer in the metadata
                if not line.startswith("#"):
                    break

                if "pydantic-sweep" in line:
                    major, minor, _ = __version__.split(".")
                    expected = rf"pydantic-sweep~={major}.{minor}.\d+"
                    match = re.search(expected, line)
                    assert match is not None, f"Version not correct: {line}"


def test_script_run():
    res = subprocess.run(
        [sys.executable, str(_EXPERIMENT_DIR / "runner.py")],
        check=True,
        capture_output=True,
        text=True,
    )
    assert "Execute main with:" in res.stdout



import enum
import runpy
import sys
import tempfile
from pathlib import Path

import pydantic
import pytest

from pydantic_sweep import BaseModel, convert
from pydantic_sweep._generation import model_to_python


class MyEnum(enum.Enum):
    a = "a"
    b = "b"


class Model(BaseModel):
    x: int = 0
    y: str = ""
    z: list = pydantic.Field(default_factory=list)
    a: set = pydantic.Field(default_factory=set)
    b: float = 0.0
    c: dict = pydantic.Field(default_factory=dict)
    d: tuple = ()
    e: Path = Path("")
    f: MyEnum = MyEnum.a


class NestedModel(BaseModel):
    sub: Model = Model()
    hidden_sub: list[Model] = pydantic.Field(default_factory=list)


class TestModelToPython:
    def _eval(self, s: str, /):
        """Write code to file, execute with runpy, and return model."""
        with tempfile.TemporaryDirectory() as tmpdir:
            model_file = Path(tmpdir) / "model.py"
            with open(model_file, "w") as f:
                f.write(s)
            res = runpy.run_path(str(model_file))
        return res["model"]

    def _test_generation(self, model: pydantic.BaseModel, /, **kwargs):
        code = model_to_python(model, **kwargs)
        model_reconstructed = self._eval(code)
        assert model_reconstructed == model
        return code

    def test_local(self):
        class Model(BaseModel):
            x: int

        with pytest.raises(ValueError):
            model_to_python(Model(x=1))

    @pytest.mark.parametrize(
        "model",
        [
            Model(x=1),
            Model(y="test"),
            Model(x=1, y="test"),
            Model(z=[1]),
            Model(a={1}),
            Model(b=1.0),
            Model(c=dict(a=1)),
            Model(d=(1,)),
            Model(e=Path("/home")),
            Model(f=MyEnum.b),
        ],
    )
    def test_basic(self, model: Model):
        self._test_generation(model)

    def test_nested(self):
        self._test_generation(NestedModel())
        self._test_generation(NestedModel(sub=Model(x=5)))

    def test_nested_hidden(self):
        self._test_generation(NestedModel(hidden_sub=[Model(x=1)]))
        self._test_generation(
            NestedModel(hidden_sub=[Model(x=1, c=dict(a=5), a={1, "a"})])
        )


@pytest.mark.parametrize("ext", ["json", "yaml", "py"])
def test_conversion(tmp_path, ext):
    submodel = Model(
        x=1,
        y="test",
        z=[1],
        a={1},
        b=1.0,
        c=dict(a=1),
        d=(1,),
        e=Path("/test"),
        f=MyEnum.b,
    )
    model = NestedModel(sub=submodel, hidden_sub=[submodel])
    if ext == "py":
        model_str = "model"
    else:
        model_str = f"{type(model).__module__}.{type(model).__name__}"

    file = tmp_path / f"model.{ext}"
    convert.write(file, model=model)
    assert convert.load(file, model=model_str) == model


def test_conversion_entrypoint(tmp_path, monkeypatch):
    """Ring-conversion between JSON, YAML, and Python files."""
    model = Model(
        x=1,
        y="test",
        z=[1],
        a={1},
        b=1.0,
        c=dict(a=1),
        d=(1,),
        e=Path("/test"),
        f=MyEnum.b,
    )
    model_str = f"{Model.__module__}.{Model.__name__}"
    json_file = tmp_path / "model.json"
    yaml_file = tmp_path / "model.yaml"
    py_file = tmp_path / "model.py"

    convert.write(tmp_path / json_file, model=model)

    modules = dict(sys.modules)
    modules.pop("pydantic_sweep.convert", None)
    monkeypatch.setattr(sys, "modules", modules)

    monkeypatch.setattr(
        "sys.argv", ["", str(json_file), str(yaml_file), "--model", model_str]
    )
    runpy.run_module(
        "pydantic_sweep.convert",
        run_name="__main__",
    )

    monkeypatch.setattr(
        "sys.argv", ["", str(yaml_file), str(py_file), "--model", model_str]
    )
    runpy.run_module(
        "pydantic_sweep.convert",
        run_name="__main__",
    )

    json_file.unlink()
    monkeypatch.setattr(
        "sys.argv", ["", str(py_file), str(json_file), "--model", "model"]
    )
    runpy.run_module(
        "pydantic_sweep.convert",
        run_name="__main__",
    )

    reconstructed_model = convert.load(json_file, model=model_str)
    assert reconstructed_model == model



import typing
from typing import Annotated, Any, Generic, TypeVar

import pydantic
import pytest
import typing_extensions
from pydantic import Discriminator, Tag, ValidationError

from pydantic_sweep._model import (
    BaseModel,
    DefaultValue,
    check_model,
    check_unique,
    config_chain,
    config_product,
    config_roundrobin,
    config_zip,
    field,
    initialize,
    model_replace,
)


class TestBaseModel:
    def test_config(self):
        class Model(BaseModel):
            x: int | float

        assert Model(x=5).x == 5

        # Wrong type for x
        with pytest.raises(ValidationError):
            Model(x=None)

        # Assign wrong type for x to instantiated model
        model = Model(x=5)
        with pytest.raises(ValidationError):
            model.x = None

        # Extra field
        with pytest.raises(ValidationError):
            Model(x=5, y=5)

        # Extra model
        with pytest.raises(ValidationError):
            Model(x=5, y=dict(x=5))

    def test_validation(self):
        class Sub1(BaseModel):
            x: pydantic.StrictInt | pydantic.StrictFloat
            y: int = 5

        class Sub2(BaseModel):
            x: pydantic.StrictInt

        class Model(BaseModel):
            sub: Annotated[Sub1 | Sub2, "Select them automatically"]

        with pytest.raises(pydantic.ValidationError):
            Model(sub=dict(x=1))

        # These should work, since they uniquely identify Sub1
        assert Model(sub=dict(x=1, y=2)) == Model(sub=Sub1(x=1, y=2))
        assert Model(sub=dict(x=1.0)) == Model(sub=Sub1(x=1.0))

        # Old-style
        class Model(BaseModel):
            sub: Sub1 | Sub2

        with pytest.raises(pydantic.ValidationError):
            Model(sub=dict(x=1))

    def test_validation_discriminator(self):
        class Sub1(BaseModel):
            x: int

        class Sub2(BaseModel):
            x: int

        class DiscModel(BaseModel):
            sub: Annotated[
                Annotated[Sub1, Tag("sub1")] | Annotated[Sub2, Tag("sub2")],
                Discriminator(lambda *args: "sub2"),
            ]

        assert DiscModel(sub=dict(x=1)) == DiscModel(sub=Sub2(x=1))

    def test_validation_dict(self):
        # This is strongly discouraged, but the validator should still works
        class Model(BaseModel):
            d: dict[str, int] | int

        Model(d=dict(x=1))

        # This is unsafe and shouldn't
        class Sub(BaseModel):
            x: int

        # Best match should prefer the pydantic Basemodel
        class Model(BaseModel):
            d: dict[str, float] | Sub

        assert Model(d=dict(x=1.0)) == Model(d=Sub(x=1.0))


class TestCheckModel:
    def test_complex(self):
        class Sub1(BaseModel):
            name: typing.Literal["Sub1"]
            x: int

        class Sub2(BaseModel):
            name: typing_extensions.Literal["Sub2"]
            x: int

        T = TypeVar("T", str, float)

        class Model(BaseModel):
            sub: Sub1 | Sub2
            x: tuple = pydantic.Field(description="Testing")
            y1: Annotated[int | Annotated[float, "s"], "Some number"]
            y2: Annotated[int | float, "Some number"]
            z: Annotated[None, "never a value"] = None
            gen: Annotated[T | float, "doc"]

        check_model(Model, unhashable="raise")

    def test_nested_fail(self):
        class A(pydantic.BaseModel):
            x: int

        class Model(BaseModel):
            x: int
            a: A

        with pytest.raises(ValueError):
            check_model(Model)
        with pytest.raises(ValueError):
            check_model(Model(x=5, a=dict(x=6)))

    def test_nested_pass(self):
        class B(BaseModel):
            x: int

        class Model2(BaseModel):
            x: int
            a: B

        check_model(Model2)
        check_model(Model2(x=5, a=dict(x=6)))

    def test_subtype(self):
        class A(pydantic.BaseModel):
            x: int

        class B(BaseModel):
            x: tuple[A]

        with pytest.raises(ValueError):
            check_model(B)

    def test_union_types(self):
        class A(BaseModel):
            x: int

        class B(BaseModel):
            x: int

        class Nested(BaseModel):
            s: A | B

        check_model(Nested)

        class Nested(BaseModel):
            s: A | pydantic.BaseModel

        with pytest.raises(ValueError):
            check_model(Nested)

    def test_arbitrary(self):
        class A(pydantic.BaseModel, extra="forbid"):
            x: int = 5

        check_model(A)
        check_model(A())

        A.model_config["arbitrary_types_allowed"] = True

        with pytest.raises(ValueError):
            check_model(A)
        with pytest.raises(ValueError):
            check_model(A())

    def test_generic(self):
        T = TypeVar("T")

        class A(BaseModel, Generic[T]):
            x: T

        # Unconstrained type variable
        with pytest.warns(UserWarning, match="`x`"):
            check_model(A)

        T = TypeVar("T", bound=str)

        class A(BaseModel, Generic[T]):
            x: T

        check_model(A)

        T = TypeVar("T", str, float)

        class A(BaseModel, Generic[T]):
            x: T

        check_model(A)

        T = TypeVar("T", bound=list[set])

        class A(BaseModel, Generic[T]):
            x: T

        with pytest.warns(UserWarning, match="`x`"):
            check_model(A)

    def test_generic_union(self):
        T = TypeVar("T")

        class Model(BaseModel):
            x: T | int

        with pytest.warns(UserWarning, match="`x`"):
            check_model(Model)

        T = TypeVar("T", str, float)

        class Model(BaseModel):
            x: T | int

        check_model(Model)

    def test_ellipsis(self):
        class Model(BaseModel):
            x: tuple[str, ...]

        check_model(Model)

    def test_any(self):
        class Model(BaseModel):
            x: tuple[Any, ...]

        with pytest.warns(UserWarning, match="`x`"):
            check_model(Model)

    def test_non_hashable(self):
        """Note: mutable types are not hashable."""

        class A(BaseModel):
            x: set

        class B(BaseModel):
            a: A

        check_model(A, unhashable="ignore")
        check_model(B, unhashable="ignore")
        with pytest.warns(UserWarning, match="`a.x`"):
            check_model(B)
        with pytest.raises(ValueError, match="`a.x`"):
            check_model(B, unhashable="raise")

        class A(BaseModel):
            y: list

        with pytest.warns(UserWarning, match="`y`"):
            check_model(A, unhashable="warn")

        class TD(typing_extensions.TypedDict):
            x: int

        class A(BaseModel):
            t: TD

        with pytest.warns(UserWarning, match="`t`"):
            check_model(A, unhashable="warn")

        class A(BaseModel):
            x: int | Annotated[set, "set"]

        with pytest.warns(UserWarning, match="`x`"):
            check_model(A)

    def test_non_hashable_nested(self):
        class A(BaseModel):
            x: tuple[tuple[list]]

        with pytest.warns(UserWarning, match="`x`"):
            check_model(A)

        class A(BaseModel):
            x: tuple[list[int]]

        with pytest.warns(UserWarning, match="`x`"):
            check_model(A)


class TestField:
    def test_invalid_path(self):
        with pytest.raises(ValueError):
            field("a-b", [1])

    def test_basic(self):
        assert field("a", []) == []
        assert field("a", [1, 2]) == [dict(a=1), dict(a=2)]
        assert field("a.b", [1]) == [dict(a=dict(b=1))]
        assert field(("a", "b"), [1]) == [dict(a=dict(b=1))]

    def test_default_value(self):
        class Model(BaseModel):
            x: int = 5

        res = initialize(Model, field("x", [1, DefaultValue, 2]))
        assert res[0].x == 1
        assert res[1].x == 5
        assert res[2].x == 2

        with pytest.raises(TypeError):
            DefaultValue()
        with pytest.raises(TypeError):

            class Test(DefaultValue):
                pass

        assert str(DefaultValue) == "DefaultValue"

    def test_check(self):
        with pytest.raises(ValueError):
            field("a", [dict(a=1)])

        res = field("a", [dict(a=1)], check=False)
        assert res == [dict(a=dict(a=1))]

    def test_iterator_values(self):
        res = field("a", iter(range(2)), check=True)
        assert res == [dict(a=0), dict(a=1)]


class TestInitialize:
    def test_basic(self):
        class Model(BaseModel):
            x: int

        assert initialize(Model, [{"x": 5}]) == [Model(x=5)]

    def test_partial(self):
        """Test partial instantiation of model."""

        class Model(BaseModel):
            x: int
            y: int = 6

        m = Model(x=100)
        m1, m2 = initialize(Model, [{"x": 10}, {"x": 11}])
        assert m1 == Model(x=10, y=6)
        assert m2 == Model(x=11, y=6)

        # Post-hoc setting of parameters
        m1.y = 10
        assert m2.y == 6
        assert m.y == 6

    def test_copy(self):
        """Make sure we do not share state between models."""

        class Sub(BaseModel):
            x: int = 5

        class Model(BaseModel):
            x: int
            sub: Sub = Sub()

        m1, m2 = initialize(Model, [{"x": 5}, {"x": 6}])
        assert m1.sub is not m2.sub
        m1.sub.x = 10
        assert m2.sub.x == 5

        m1, m2 = initialize(Model, [{"x": 5}, {"x": 6}], default=dict(x=10))
        assert m1.sub is not m2.sub
        m1.sub.x = 10
        assert m2.sub.x == 5

    def test_default(self):
        class Sub(BaseModel):
            x: int

        class Model(BaseModel):
            sub: Sub
            y: int = 0

        res = initialize(Model, field("sub.x", [1]), default=dict(y=10))
        assert res == [Model(sub=Sub(x=1), y=10)]

        # Default as default value should not have any effect
        res = initialize(Model, field("sub.x", [1]), default=dict(y=DefaultValue))
        assert res == [Model(sub=Sub(x=1), y=0)]

        res = initialize(
            Model,
            field("sub.x", [DefaultValue]),
            constant=dict(y=10),
            default={"sub.x": 99},
        )
        assert res == [Model(sub=Sub(x=99), y=10)]

        res = initialize(
            Model,
            field("sub.x", [DefaultValue]),
            constant=dict(y=10),
            default=dict(sub=dict(x=99)),
        )
        assert res == [Model(sub=Sub(x=99), y=10)]

    def test_constant(self):
        class Sub(BaseModel):
            x: int

        class Model(BaseModel):
            sub: Sub
            y: int = 0

        res = initialize(Model, field("sub.x", [1]), constant=dict(y=10))
        assert res == [Model(sub=Sub(x=1), y=10)]

        res = initialize(Model, field("sub.x", [1]), constant=dict(y=DefaultValue))
        assert res == [Model(sub=Sub(x=1), y=0)]

        res = initialize(Model, field("sub.x", [1, 2]), constant=dict(y=10))
        assert res == [Model(sub=Sub(x=1), y=10), Model(sub=Sub(x=2), y=10)]

        res = initialize(Model, [dict()], constant={"sub.x": 0})
        assert res == [Model(sub=Sub(x=0))]

        # Provide nested diction config for constant
        res = initialize(Model, [dict()], constant=dict(sub=dict(x=0)))
        assert res == [Model(sub=Sub(x=0))]

        with pytest.raises(TypeError):
            initialize(Model, [], constant=[5])

        with pytest.raises(ValueError):
            initialize(Model, field("sub.x", [1]), constant={"sub.x": 2})

        # Also DefaultValue should conflict here
        with pytest.raises(ValueError):
            initialize(Model, field("sub.x", [1]), constant={"sub.x": DefaultValue})

    def test_to(self):
        class Sub(BaseModel):
            x: int = 5

        values = field("x", [0, 1])
        assert initialize(Sub, values, to="a.b") == field(
            "a.b", initialize(Sub, values)
        )

        sub = initialize(Sub, field("x", [0]), to="s")
        assert sub == field("s", [Sub(x=0)])

        class Model(BaseModel):
            s: Sub

        model = initialize(Model, sub)
        assert model == [Model(s=Sub(x=0))]

    def test_at(self):
        class Sub(BaseModel):
            x: int

        class Model(BaseModel):
            sub: Sub

        values = field("sub.x", [0])
        partial = initialize(Sub, values, at="sub")
        assert partial == [dict(sub=Sub(x=0))]

    def test_conflicing_args(self):
        class Sub(BaseModel):
            x: int

        initialize(Sub, [dict(x=1)])
        with pytest.raises(ValueError):
            initialize(Sub, [dict(x=1)], at="sub", to="sub")


def test_config_product():
    res = config_product(field("a", [1, 2]), field("b", [3, 4]))
    expected = [dict(a=1, b=3), dict(a=1, b=4), dict(a=2, b=3), dict(a=2, b=4)]
    assert res == expected

    with pytest.raises(ValueError):
        config_product(field("a", [1]), field("a", [2]))


def test_config_zip():
    res = config_zip(field("a", [1, 2]), field("b", [3, 4]))
    assert res == [dict(a=1, b=3), dict(a=2, b=4)]

    # Different lengths
    with pytest.raises(ValueError):
        config_zip(field("a", [1, 2]), field("b", [3]))

    # Same value
    with pytest.raises(ValueError):
        config_zip(field("a", [1, 2]), field("a", [3, 4]))


def test_config_chain():
    res = config_chain(field("a", [1, 2]), field("b", [3, 4]))
    assert res == [dict(a=1), dict(a=2), dict(b=3), dict(b=4)]

    # Same keys should not cause conflicts here
    res = config_chain(field("a", [1]), field("b", [2]))
    assert res == [dict(a=1), dict(b=2)]


def test_config_roundrobin():
    res = config_roundrobin(field("a", [1, 2]), field("b", [3, 4]))
    assert res == [dict(a=1), dict(b=3), dict(a=2), dict(b=4)]

    # Same keys should not cause conflicts here
    res = config_chain(field("a", [1]), field("b", [2]))
    assert res == [dict(a=1), dict(b=2)]


def test_default_override():
    """Make sure default values cannot be overwritten."""
    with pytest.raises(ValueError):
        config_product(
            field("a", [DefaultValue]),
            field("a", [DefaultValue]),
        )

    with pytest.raises(ValueError):
        config_product(
            field("a", [1.0]),
            field("a", [DefaultValue]),
        )


def test_check_unique():
    xs = field("a", [1, 2])
    assert check_unique(xs, raise_exception=False)
    assert not check_unique(xs, xs, raise_exception=False)
    assert not check_unique([*xs, *xs], raise_exception=False)

    class Sub(BaseModel):
        x: int = 0

    class Model(BaseModel):
        sub: Sub = Sub()
        y: int = 0

    assert check_unique(
        [Model(), Model(sub=Sub(x=1)), Model(y=1)], raise_exception=False
    )
    assert not check_unique(
        [Model(y=1), Model(sub=Sub(x=1)), Model(y=1)], raise_exception=False
    )

    with pytest.raises(ValueError):
        check_unique([Model(y=0), Model(y=0)])
    check_unique([Model(y=0)])

    class Sub2(Sub):
        pass

    # These should be marked as unique, since they are different model classes
    check_unique([Model(sub=Sub()), Model(sub=Sub2())])

    # Check individual models
    check_unique(Model(y=0))
    check_unique(Model(y=0), Model(y=1))
    with pytest.raises(ValueError):
        check_unique(Model(), Model())


class TestModelReplace:
    def test_basic(self):
        class M(BaseModel):
            x: int = 0
            y: int = 0

        class B(BaseModel):
            m: M = M()
            z: int = 0

        b = B(m=M(x=1, y=2), z=3)
        b1 = model_replace(b, values={})
        assert b1 == b
        assert b1 is not b

        b1 = model_replace(b, values={"m.x": 99})
        assert b1 == B(m=M(x=99, y=2), z=3)
        b1 = model_replace(b, values={("m", "x"): 99})
        assert b1 == B(m=M(x=99, y=2), z=3)
        b1 = model_replace(b, values=dict(m=dict(x=99)))
        assert b1 == B(m=M(x=99, y=2), z=3)
        b1 = model_replace(b, values=dict(m=dict(x=99, y=98)))
        assert b1 == B(m=M(x=99, y=98), z=3)
        b1 = model_replace(b, values=dict(m=M(x=99)))
        assert b1 == B(m=M(x=99), z=3)

    def test_errors(self):
        class M(BaseModel):
            x: int = 0
            y: int = 0

        class B(BaseModel):
            m: M = M()

        class A(BaseModel):
            b: B = B()
            z: int = 0

        a = A(b=dict(m=M(x=1, y=2)), z=3)
        # Providing conflicting values
        with pytest.raises(ValueError):
            model_replace(a, values={"b.m.x": 99, "b.m": dict(x=98)})

        # Extra fields should not be allowed
        with pytest.raises(ValueError):
            model_replace(a, values={"b.m.z": 99})

    def test_default(self):
        class A(BaseModel):
            x: int = 0
            y: int = 0

        class B(BaseModel):
            a: A = A()

        a = A(x=1, y=2)
        a1 = model_replace(a, values={"x": DefaultValue})
        assert a1 == A(x=0, y=a.y)
        a1 = model_replace(a, values={"x": DefaultValue, "y": DefaultValue})
        assert a1 == A()

        # Deeper nesting
        b = B(a=a)
        b1 = model_replace(b, values={"a.x": DefaultValue})
        assert b1 == B(a=A(x=0, y=b.a.y))

        b = B()
        b1 = model_replace(b, values={"a.x": DefaultValue})
        assert b1 == b
        assert b1 is not b

        # conflicts
        b = B(a=A(x=1, y=2))
        b1 = model_replace(b, values={"a.x": DefaultValue, "a": DefaultValue})
        assert b1 == B()
        b1 = model_replace(b, values={"a": DefaultValue, "a.x": DefaultValue})
        assert b1 == B()



from typing import Literal

import pydantic

import pydantic_sweep as ps


class Method(ps.BaseModel):
    optimizer: Literal["SGD", "Adam"] = "Adam"
    lr: float = 3e-4


class ExperimentConfig(ps.BaseModel):
    seed: int | None = None
    method: Method = pydantic.Field(default_factory=Method)



from __future__ import annotations

import contextlib
from collections.abc import Collection, Generator, Mapping
from enum import Enum
from pathlib import Path
from typing import Any, cast

import pydantic

__all__ = ["model_to_python"]


def model_to_python(
    model: pydantic.BaseModel,
    *,
    name: str = "model",
    exclude_unset: bool = True,
    exclude_defaults: bool = False,
    include: Collection[str] | Mapping[str, Any] | None = None,
) -> str:
    """Generate python code for a pydantic model.

    This function generates python code that instantiates a given model. This is,
    for example, useful to switch json/yaml configuration files to Python-native ones.

    Parameters
    ----------
    model :
        The model that we want to convert to Python code
    name :
        The name of the variable to which we assign the instantiated model.
    exclude_defaults :
        Whether to exclude default arguments.
    include :
        Additional fields to include.

    Returns
    -------
    The corresponding python code including imports.
    """
    model_classes: set[type[object]] = set()
    lines: list[str] = []
    include = cast(None | dict | set[str], include)  # pydantic is too strict here
    dump = model.model_dump(
        exclude_defaults=exclude_defaults,
        exclude_unset=exclude_unset,
        include=include,
    )

    _add_python_code(
        model=model,
        field_prefix=f"{name} = ",
        dump=dump,
        indent=0,
        lines=lines,
        model_classes=model_classes,
    )

    import_lines = _generate_imports(model_classes)
    return "\n".join([*sorted(import_lines), "", "", *lines])


def _generate_imports(classes: Collection[type[object]], /) -> list[str]:
    """Generate import statements for the given classes."""
    if len(set(cls.__name__ for cls in classes)) != len(classes):
        from collections import Counter

        counts = Counter(cls.__name__ for cls in classes)
        duplicates = [name for name, count in counts.items() if count > 1]
        raise ValueError(
            "The following models share the same name, but exist at different code "
            f"paths. This is currently not supported: {', '.join(duplicates)}."
        )

    return [f"from {cls.__module__} import {cls.__name__}" for cls in classes]


def _add_import(obj: Any, model_classes: set) -> None | str:
    cls = type(obj)
    if cls.__qualname__ != cls.__name__:
        raise ValueError("Cannot generate code for local modules.")
    if cls.__module__ == "builtins":
        return None
    else:
        model_classes.add(cls)
        return cls.__name__


@contextlib.contextmanager
def _wrapped(
    whitespace: str, prefix: str, open: str, close: str, lines: list[str]
) -> Generator[None, None, None]:
    lines.append(f"{whitespace}{prefix}{open}")
    yield
    # Final line comma not needed (works better with auto-formatting)
    lines[-1] = lines[-1].removesuffix(",")
    if whitespace:
        close += ","
    lines.append(f"{whitespace}{close}")


def _add_python_code(
    model: Any,
    *,
    field_prefix: str,
    dump: Any,
    indent: int,
    lines: list,
    model_classes: set,
) -> None:
    whitespace = "    " * indent

    if isinstance(model, pydantic.BaseModel):
        type_name = _add_import(model, model_classes)
        assert type_name is not None
        assert isinstance(dump, dict)
        open, close = f"{type_name}(", ")"
        with _wrapped(whitespace, field_prefix, open, close, lines):
            for key, dump_value in dump.items():
                _add_python_code(
                    model=getattr(model, key),
                    field_prefix=f"{key}=",
                    dump=dump_value,
                    lines=lines,
                    model_classes=model_classes,
                    indent=indent + 1,
                )
    elif isinstance(model, list | tuple | set):
        type_name = _add_import(model, model_classes=model_classes)

        if isinstance(model, list):
            open, close = "[", "]"
        elif isinstance(model, tuple):
            open, close = "[", "]"
        elif isinstance(model, set):
            open, close = "{", "}"
        else:
            raise ValueError(f"Unsupported collection type: {type(model)}")

        if type_name is not None:
            open = f"{type_name}{open}"
            close = f"{close})"

        with _wrapped(whitespace, field_prefix, open, close, lines):
            for model_item, dump_item in zip(model, dump, strict=True):
                _add_python_code(
                    model=model_item,
                    field_prefix="",
                    dump=dump_item,
                    lines=lines,
                    model_classes=model_classes,
                    indent=indent + 1,
                )
    elif isinstance(model, dict):
        type_name = _add_import(model, model_classes=model_classes)
        if type_name is None:
            open, close = "{", "}"
        else:
            open = f"{type_name}({{"
            close = "})"

        with _wrapped(whitespace, field_prefix, open, close, lines):
            for key, dump_value in dump.items():
                _add_python_code(
                    model=model[key],
                    field_prefix=f"{key!r}: ",
                    dump=dump_value,
                    lines=lines,
                    model_classes=model_classes,
                    indent=indent + 1,
                )
    elif isinstance(dump, Enum):
        enum_cls = _add_import(dump, model_classes=model_classes)
        dump = f"{enum_cls}.{dump.name}"
        lines.append(f"{whitespace}{field_prefix}{dump},")
    else:
        if isinstance(dump, Path):
            dump = str(dump)
        lines.append(f"{whitespace}{field_prefix}{dump!r},")



from . import types
from ._model import (
    BaseModel,
    DefaultValue,
    check_model,
    check_unique,
    config_chain,
    config_combine,
    config_product,
    config_roundrobin,
    config_zip,
    field,
    initialize,
    model_replace,
)
from ._utils import as_hashable, model_diff, random_seeds
from ._version import __version__

__all__ = [
    "BaseModel",
    "DefaultValue",
    "__version__",
    "as_hashable",
    "check_model",
    "check_unique",
    "config_chain",
    "config_combine",
    "config_product",
    "config_roundrobin",
    "config_zip",
    "field",
    "initialize",
    "model_diff",
    "model_replace",
    "random_seeds",
    "types",
]



from __future__ import annotations

import itertools
import types
import typing
from collections.abc import Hashable, Iterable
from functools import partial
from typing import Any, Literal, TypeVar, overload

import more_itertools
import pydantic

from pydantic_sweep._utils import (
    _flexible_config_to_nested,
    as_hashable,
    items_skip,
    iter_subtypes,
    merge_nested_dicts,
    nested_dict_at,
    nested_dict_from_items,
    nested_dict_get,
    nested_dict_items,
    nested_dict_replace,
    normalize_path,
    notebook_link,
    path_to_str,
    raise_warn_ignore,
)
from pydantic_sweep.types import (
    BaseModelT,
    Chainer,
    Combiner,
    Config,
    FieldValue,
    FlexibleConfig,
    Path,
    StrictPath,
)

__all__ = [
    "BaseModel",
    "DefaultValue",
    "check_model",
    "check_unique",
    "config_chain",
    "config_combine",
    "config_product",
    "config_roundrobin",
    "config_zip",
    "field",
    "initialize",
    "model_replace",
]

T = TypeVar("T")


class BaseModel(
    pydantic.BaseModel,
    extra="forbid",
    validate_assignment=True,
    arbitrary_types_allowed=False,
    validate_return=True,
):
    """Base model with validation enabled by default."""

    @pydantic.model_validator(mode="before")
    @classmethod
    def _safe_union_validator(cls, data: Any) -> Any:
        """Disallow unsafe matches to nested Union models.

        By default, pydantic does not raise an error if multiple pydantic models in a
        union type could match the provided data.
        """
        if isinstance(data, dict):
            fields = cls.model_fields
            for key, value in data.items():
                # Only dicts needs special handling, since static values cannot
                # represent nested models. This also covers the case if value is
                # already a pydantic model.
                if not isinstance(value, dict):
                    continue

                # extra items are handled by extra='forbid' model setting.
                try:
                    field = fields[key]
                except KeyError:
                    continue

                # Discriminators are an alternative way to handle this
                if field.discriminator is not None or any(
                    isinstance(m, pydantic.Discriminator) for m in field.metadata
                ):
                    continue

                # We focus on direct unions for now, since they will be the most
                # common use-case. This does not check things like `tuple[Sub1 | Sub2]`.
                origin = typing.get_origin(field.annotation)
                if origin not in (typing.Union, types.UnionType):
                    continue

                # Manually validate each model.
                matches = []
                for annotation in typing.get_args(field.annotation):
                    # Any other type should not need validation, since either they
                    # can't match or, in the case of dictionaries, pydantic models
                    # are preferred under the best-match strategy.
                    if not isinstance(annotation, type):
                        continue
                    try:
                        issub = issubclass(annotation, pydantic.BaseModel)
                    except TypeError:
                        continue

                    if issub:
                        try:
                            res = annotation.model_validate(value)  # type: ignore[attr-defined]
                        except pydantic.ValidationError:
                            pass
                        else:
                            matches.append((annotation.__name__, res))

                if len(matches) > 1:
                    from pydantic_core import PydanticCustomError

                    raise PydanticCustomError(
                        "unsafe_union_error",
                        "Multiple models of a Union type could match the provided "
                        "data: {conflicts}. To avoid this error, either "
                        "initialize the nested model manually using the `initialize` "
                        "method or use a discriminated union. See {docs} for details.",
                        dict(
                            conflicts=", ".join([name for name, _ in matches]),
                            docs=notebook_link("nested"),
                        ),
                    )
                elif matches:
                    # Avoid re-running the model validation downstream
                    data[key] = matches[0][1]

        return data


class NameMetaClass(type):
    """A metaclass that overwrite cls.__str__ to its name"""

    def __str__(cls) -> str:
        return cls.__name__


class DefaultValue(metaclass=NameMetaClass):
    """Indicator class for a default value in the ``field`` method."""

    def __new__(cls, *args: Any, **kwargs: Any) -> DefaultValue:
        raise TypeError("This is a sentinel value and not meant to be instantiated.")

    def __init_subclass__(cls, **kwargs: Any) -> None:
        raise TypeError("This is a sentinel value and not meant to be subclassed.")


def _field_str(t: Any, /, *, path: StrictPath) -> str:
    """Field and type info at a given path.

    >>> _field_str(5., path=())
    'float'

    >>> _field_str(5., path=("a", "b"))
    'float at field `a.b`'
    """
    name = t.__name__ if hasattr(t, "__name__") else t.__class__.__name__
    field_msg = f" at field `{path_to_str(path)}`" if path else ""
    return f"{name}{field_msg}"


def _check_model_config(
    model: pydantic.BaseModel | type[pydantic.BaseModel], /, *, path: StrictPath
) -> None:
    config = model.model_config
    if "extra" not in config or config["extra"] != "forbid":
        info = _field_str(model, path=path)
        raise ValueError(
            f"Model {info} must have 'extra=forbid' option enabled. "
            "Without this, typos in field names will be silently ignored."
        )
    if config.get("arbitrary_types_allowed", False):
        info = _field_str(model, path=path)
        raise ValueError(
            f"Model {info} must have 'arbitrary_types_allowed=False' set. "
            "Configuration classes should be built with basic types only to ensure "
            "that the configuration can be checked and serialized reliably."
        )


def check_model(
    model: pydantic.BaseModel | type[pydantic.BaseModel],
    /,
    *,
    unhashable: Literal["warn", "ignore", "raise"] = "warn",
) -> None:
    """Best-effort check that the model has the correct configuration.

    This recurses into the models, but there's probably a way to achieve a
    false positive if one tries.

    Parameters
    ----------
    model :
        The model to check.
    unhashable :
        The action to take when a non-hashable type hint is encountered in the mode.
    """
    to_check: list[tuple[StrictPath, Any]] = [((), model)]
    checked = set()

    while to_check:
        path, model = to_check.pop()

        if isinstance(model, pydantic.BaseModel):
            name = model.__class__.__name__
        # Subclass can raise error for inputs that are not type
        # https://github.com/python/cpython/issues/101162
        elif isinstance(model, type) and issubclass(model, pydantic.BaseModel):
            name = model.__name__
        else:
            # Just a leaf node
            if isinstance(model, type) and not issubclass(model, Hashable):
                info = _field_str(model, path=path)
                raise_warn_ignore(
                    f"Non-hashable type {info}. These can lead to accidental "
                    f"shared state between different configuration objects.",
                    action=unhashable,
                )
            # Quirk: typing.Any is Hashable
            elif model is typing.Any:
                field = path_to_str(path)
                raise_warn_ignore(
                    f"Unconstrained variable (type Any) at field `{field}`. These "
                    f"cannot be validated by pydantic and are thus not suitable for "
                    f"configuration classes.",
                    action=unhashable,
                )
            continue

        if name in checked:
            continue

        _check_model_config(model, path=path)
        checked.add(name)

        cls = type(model) if isinstance(model, pydantic.BaseModel) else model
        for name, field in cls.model_fields.items():
            annotation = field.annotation
            if annotation is not None:
                for sub_type in iter_subtypes(annotation):
                    to_check.append(((*path, name), sub_type))  # noqa: PERF401


def _config_prune_default(config: Config) -> Config:
    """Prune default value placeholders from a config.

    This allows pydantic to handle initialization of them.
    """
    items = nested_dict_items(config)
    items = items_skip(items, target=DefaultValue)
    return nested_dict_from_items(items)


@overload
def initialize(
    model: type[BaseModelT],
    configs: Iterable[Config],
    *,
    constant: FlexibleConfig | None = None,
    default: FlexibleConfig | None = None,
    to: Path,
    at: Path | None = None,
    check: bool = True,
) -> list[Config]:
    pass


@overload
def initialize(
    model: type[BaseModelT],
    configs: Iterable[Config],
    *,
    constant: FlexibleConfig | None = None,
    default: FlexibleConfig | None = None,
    to: Path | None = None,
    at: Path,
    check: bool = True,
) -> list[Config]:
    pass


@overload
def initialize(
    model: type[BaseModelT],
    configs: Iterable[Config],
    *,
    constant: FlexibleConfig | None = None,
    default: FlexibleConfig | None = None,
    to: None = None,
    at: None = None,
    check: bool = True,
) -> list[BaseModelT]:
    pass


def initialize(
    model: type[BaseModelT],
    configs: Iterable[Config],
    *,
    constant: FlexibleConfig | None = None,
    default: FlexibleConfig | None = None,
    to: Path | None = None,
    at: Path | None = None,
    check: bool = True,
) -> list[Config] | list[BaseModelT]:
    """Instantiate the models with the given parameters.

    Parameters
    ----------
    model:
        The pydantic model that we want to finalize. This can be either a model cass
        or an instance of a specific model. In both cases, the configuration is checked
        for safety and the models are instantiated.
    configs:
        The partial config dictionaries that we want to initialize with pydantic.
    constant:
        Constant values that should be initialized for all models. These are safely
        merged with the parameters. Can be either a nested, or a flattened dictionary.
    default:
        Default parameter that are initialized for all models, but may be overwritten by
        other fields without any error checking. Can be either a nested or a flattened
        dictionary.
    to:
        If provided, will first initialize the model and then return a
        configuration dictionary that sets the model as the values at the given path.
        Essentially a shortcut to first passing the models to ``field(to, models)``.
    at:
        If provided, will initialize the model at the given path in the configuration.
    check:
        Whether to apply error checks to model.
    """
    if check:
        # Unhashable are not a deal-breaker on the model side, and we deal with them
        # already on the parameter side.
        check_model(model, unhashable="ignore")

    if constant is not None:
        if not isinstance(constant, dict):
            raise TypeError(
                f"Expected dictionary for input 'constant', got '{type(constant)}'."
            )
        configs = config_product(configs, [_flexible_config_to_nested(constant)])

    # Remove placeholders now
    configs = [_config_prune_default(config) for config in configs]

    if default is not None:
        if not isinstance(default, dict):
            raise TypeError(
                f"Expected dictionary for input 'default', got '{type(default)}'."
            )
        # A DefaultValue as a default should not change anything
        default_config = _flexible_config_to_nested(default, skip=DefaultValue)
        configs = [
            merge_nested_dicts(default_config, param, overwrite=True)
            for param in configs
        ]

    # Initialize a subconfiguration at the path ``at``
    if at is not None:
        if to is not None:
            raise ValueError("Only on of `path` and `at` can be provided, not both.")

        subconfigs = [nested_dict_get(param, at, leaf=False) for param in configs]
        submodels = initialize(model, subconfigs)
        return [
            nested_dict_replace(param, path=at, value=submodel)
            for param, submodel in zip(configs, submodels)
        ]

    # Initialize the provided models
    models = [model(**config) for config in configs]

    if to is not None:
        # Check not needed here: values are all pydantic.BaseModel by design
        return field(to, models, check=False)
    else:
        return models


def model_replace(model: BaseModelT, *, values: FlexibleConfig) -> BaseModelT:
    """Create a copy of the pydantic model with nested value replacement.

    Parameters
    ----------
    model :
        The pydantic model to replace the fields in.
    values :
        A dictionary of fields to overwrite in the model. This can be either a nested
        dictionary or a flat dictionary. A `DefaultValue` in the dictionary means that
        the field will take the default value from the base model.

    Returns
    -------
    BaseModelT:
        A new instance of the model with the fields replaced.
    """
    if not isinstance(values, dict):
        raise TypeError(
            f"Expected dictionary for input 'values', got '{type(values)}'."
        )

    model_dump = model.model_dump(exclude_defaults=True)

    # We remove any paths with DefaultValue both from the model dump and values,
    # so that pydantic uses whatever was the default.
    config_items = []
    for key, value in nested_dict_items(values):
        if value is DefaultValue:
            d = model_dump
            *path, final = key
            for subkey in path:
                if subkey not in d:
                    break
                d = d[subkey]
            if final in d:
                del d[final]
        else:
            config_items.append((key, value))
    config_values = nested_dict_from_items(config_items)

    merged_config = merge_nested_dicts(model_dump, config_values, overwrite=True)
    return model.model_validate(merged_config)


def field(
    path: Path, /, values: Iterable[FieldValue], *, check: bool = True
) -> list[Config]:
    """Assign various values to a field in a pydantic Model.

    Parameters
    ----------
    path :
        The path to the key in the model. Can either be a dot-separated string of
        keys (e.g., ``my.key``) or a tuple of keys (e.g., ``('my', 'key')``.
    values :
        The different values that should be assigned to the field. Note that the
        `DefaultValue` class has a special meaning, since it will be effectively
        ignored, allowing it to be kept to the default model.
    check
        If ``True``, check that values are indeed hashable or pydantic Models.

    Returns
    -------
    list[Config]:
        A list of partial configuration dictionaries that can be passed to the pydantic
        model.

    Examples
    --------
    >>> import pydantic_sweep as ps

    >>> class Sub(ps.BaseModel):
    ...     x: int = 5
    ...     y: int = 6

    >>> class Model(ps.BaseModel):
    ...     sub: Sub
    ...     seed: int = 5

    >>> _ = Model.model_rebuild()

    >>> configs = ps.field("sub.x", [10, 20])
    >>> ps.initialize(Model, configs)
    [Model(sub=Sub(x=10, y=6), seed=5), Model(sub=Sub(x=20, y=6), seed=5)]

    """
    path = normalize_path(path, check_keys=True)
    if isinstance(values, str):
        raise ValueError("values must be iterable, but got a string")

    if check:
        # Iterators may get exhausted
        values = list(values)
        for value in values:
            # Note: DefaultValue is hashable
            if not isinstance(value, pydantic.BaseModel | Hashable):
                raise ValueError(
                    f"Value {value} of type {type(value)} is not hashable, which can "
                    f"cause unexpected behaviors. You can disable this check by "
                    f"passing `check=False` as a keyword argument."
                )

    return [nested_dict_at(path, value) for value in values]


def config_combine(
    *configs: Iterable[Config],
    combiner: Combiner | None = None,
    chainer: Chainer | None = None,
) -> list[Config]:
    """Flexible combination of configuration dictionaries.

    In contrast to the more specific functions below, this allows you to flexibly use
    existing functions from ``itertools`` in order to create new combiners. All
    existing combiners build on top of this function.

    The output of this function is a valid input to both itself and other combiner
    functions.

    Parameters
    ----------
    configs :
        The configurations we want to combine.
    combiner :
        A function that takes as input multiple iterables and yields tuples.
        For example: ``itertools.product``.
    chainer :
        A function that takes as input multiple iterables and yields a single new
        iterable. For example: ``itertools.chain``.

    Returns
    -------
    list[Config]:
        A list of new configuration objects after combining or chaining.
    """
    if combiner is not None:
        if chainer is not None:
            raise ValueError("Can only provide `combiner` or `chainer`, not both")
        return [merge_nested_dicts(*combo) for combo in combiner(*configs)]
    elif chainer is not None:
        res = list(chainer(*configs))
        if not isinstance(res[0], dict):
            raise ValueError(
                f"Chained items are not dictionaries, but {type(res[0])}. Are you sure "
                f"that you passed a valid chainer function? "
            )
        return res
    else:
        raise ValueError("Must provide one of `single_out` or `multi_out`")


def config_product(*configs: Iterable[Config]) -> list[Config]:
    """A product of existing configuration dictionaries.

    This is the most common way of constructing searches. It constructs the product
    of the inputs.

    >>> config_product(field("a", [1, 2]), field("b", [3, 4]))
    [{'a': 1, 'b': 3}, {'a': 1, 'b': 4}, {'a': 2, 'b': 3}, {'a': 2, 'b': 4}]

    The output of this function is a valid input to both itself and other combiner
    functions.
    """
    return config_combine(*configs, combiner=itertools.product)


def config_zip(*configs: Iterable[Config]) -> list[Config]:
    """Return the zip-combination of configuration dictionaries.

    >>> config_zip(field("a", [1, 2]), field("b", [3, 4]))
    [{'a': 1, 'b': 3}, {'a': 2, 'b': 4}]
    """
    safe_zip = partial(zip, strict=True)
    return config_combine(*configs, combiner=safe_zip)


def config_chain(*configs: Iterable[Config]) -> list[Config]:
    """Chain configuration dictionaries behind each other.

    >>> config_chain(field("a", [1, 2]), field("b", [3, 4]))
    [{'a': 1}, {'a': 2}, {'b': 3}, {'b': 4}]
    """
    return config_combine(*configs, chainer=itertools.chain)


def config_roundrobin(*configs: Iterable[Config]) -> list[Config]:
    """Interleave the configuration dictionaries.

    This is the same behavior as `config_chain`, but instead of chaining them behind
    each other, takes from the different iterables in turn.

    >>> config_roundrobin(field("a", [1, 2, 3]), field("b", [3, 4]))
    [{'a': 1}, {'b': 3}, {'a': 2}, {'b': 4}, {'a': 3}]
    """
    return config_combine(*configs, chainer=more_itertools.roundrobin)


def check_unique(
    *models_: Config | pydantic.BaseModel | Iterable[Config | pydantic.BaseModel],
    raise_exception: bool = True,
) -> bool:
    """Check that models are unique.

    Parameters
    ----------
    *models_
        Iterables of models to check for uniqueness. If multiple are passed, they are
        chained together and jointly checked.
    raise_exception :
        If ``False``, return a boolean instead of raising an exception on failure.

    Raises
    ------
    ValueError
        If models are not unique.
    """
    seen = set()
    for models in models_:
        if isinstance(models, pydantic.BaseModel | dict):
            models = [models]
        for model in models:
            model_hash = hash(as_hashable(model))
            if model_hash in seen:
                if raise_exception:
                    raise ValueError(f"The following model is not unique: {model}.")
                else:
                    return False
            seen.add(model_hash)

    return True



# /// script
# dependencies = [
#   "pydantic-sweep~=0.3.1"
# ]
# ///

import subprocess
import sys
from pathlib import Path

from config import ExperimentConfig

import pydantic_sweep as ps

# Construct experiment configurations
experiments = ps.initialize(
    ExperimentConfig,
    ps.config_product(
        ps.field("seed", ps.random_seeds(3)),
        ps.config_zip(
            ps.field("method.optimizer", ["SGD", "Adam", ps.DefaultValue]),
            ps.field("method.lr", [1e-6, 1e-4, ps.DefaultValue]),
        ),
    ),
)

# Make sure they are unique
ps.check_unique(experiments)

# Call the script with all experiment configurations
script = Path(__file__).parent / "train.py"
for experiment in experiments:
    # Here, were calling subprocess with the current python executable. On a cluster,
    # one would instead schedule the corresponding run.
    subprocess.run(
        [sys.executable, script, "--json", experiment.model_dump_json()],
        check=True,
    )



import argparse
import importlib
import os
import runpy
from pathlib import Path
from typing import Any

import pydantic
from typing_extensions import Self

from pydantic_sweep._generation import model_to_python

__all__ = ["load", "model_to_python", "write"]


def _import_module(name: str, /) -> Any:
    """Import a module by name."""
    module_name, class_name = name.rsplit(".", maxsplit=1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def load(source: os.PathLike | str, /, *, model: str) -> pydantic.BaseModel:
    """Load a pydantic model from a file.

    Parameters
    ----------
    source :
        The file that we want to read the model from. Supported extensions are
        ``.json``, ``.yaml`` and ``.py``.
    model :
        The model that we want to load. This is either a path to a module with
        dot-notation (e.g., ``my_module.MyModel``) or, for python files, the
        name of the variable where the model is stored in the file.

    Returns
    -------
    The loaded pydantic model.
    """
    source = Path(source)

    if source.suffix == ".json":
        with source.open("r") as file:
            json_str = file.read()
        cls = _import_module(model)
        return cls.model_validate_json(json_str)
    elif source.suffix == ".yaml":
        import yaml  # type: ignore[import-untyped]

        with source.open("r") as file:
            content = yaml.safe_load(file)

        cls = _import_module(model)
        return cls(**content)
    elif source.suffix == ".py":
        return runpy.run_path(str(source))[model]
    else:
        raise ValueError(f"Unsupported file type: {source.suffix}")


def write(
    target: os.PathLike | str,
    /,
    *,
    model: pydantic.BaseModel,
    yaml_options: dict[str, Any] | None = None,
    exclude_unset: bool = True,
    exclude_defaults: bool = False,
) -> None:
    """Write a model to a file.

    Parameters
    ----------
    target :
        The file to write to. Supported extensions are ``.json``, ``.yaml`` and
        ``.py``.
    model :
        The pydantic model that we want to write to the file.
    yaml_options :
        Optional keyword-arguments passed to ``yaml.dump``.
    exclude_unset :
        Exclude fields that don't have a value set.
    exclude_defaults :
        Exclude fields that have the default value set.
    """
    target = Path(target)
    if target.exists():
        raise FileExistsError(f"File already exists: {target}")
    if target.suffix == ".json":
        dump = model.model_dump_json(
            exclude_unset=exclude_unset, exclude_defaults=exclude_defaults
        )
        with target.open("w") as f:
            f.write(dump)
    elif target.suffix == ".yaml":
        import json

        import yaml

        if yaml_options is None:
            yaml_options = dict()
        # This runs serializers properly (e.g., for Path / Enum objects)
        dump = json.loads(
            model.model_dump_json(
                exclude_unset=exclude_unset, exclude_defaults=exclude_defaults
            )
        )
        with target.open("w") as f:
            yaml.dump(dump, f, **yaml_options)
    elif target.suffix == ".py":
        code = model_to_python(
            model, exclude_unset=exclude_unset, exclude_defaults=exclude_defaults
        )
        with target.open("w") as f:
            f.write(code)
    else:
        raise ValueError(f"Unsupported file type: {target.suffix}")


class Config(pydantic.BaseModel, extra="forbid"):
    """Command line configuratio nobject"""

    source: Path
    target: Path
    model: str
    yaml_default_flow_style: bool
    exclude_defaults: bool = False
    include_unset: bool = False

    @classmethod
    def from_cli(cls) -> Self:
        parser = argparse.ArgumentParser(
            "Convert Pydantic models between different python, json and yaml."
        )
        parser.add_argument("source", type=str, help="The file to convert from.")
        parser.add_argument("target", type=str, help="The file to convert to.")
        parser.add_argument(
            "--model",
            type=str,
            default=None,
            help=(
                "The model: Either the dot-separated import path for the pydantic "
                "Model that we want to load or, for python files, the name of the "
                "variable where the model is stored inside of the file."
            ),
        )
        parser.add_argument(
            "--exclude-defaults",
            action="store_true",
            help="Exclude fields that have default values.",
        )
        parser.add_argument(
            "--include-unset",
            action="store_true",
            help="Also include fields that have no values set.",
        )
        parser.add_argument(
            "--yaml-default-flow-style",
            action="store_true",
            help="Use the default flow style for yaml.",
        )
        return cls(**vars(parser.parse_args()))


if __name__ == "__main__":
    config = Config.from_cli()
    model = load(config.source, model=config.model)
    write(
        config.target,
        model=model,
        exclude_defaults=config.exclude_defaults,
        exclude_unset=not config.include_unset,
        yaml_options={"default_flow_style": config.yaml_default_flow_style},
    )
